/**
 * Copyright (c) 2008-2012 EBMWebsourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.jbidescriptor;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.net.URL;

import javax.xml.XMLConstants;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;

/**
 * CDK JBI descriptor Builder. Provide marshalling/unmarshalling features.
 * @author Christophe HAMERLING - EBM WebSourcing
 * @author Roland NAUDIN - EBM WebSourcing
 */
public class JBIDescriptorBuilder {

    /**
     * CDK schema
     */
    public static final String CDKJBI_XSD = "CDKjbi.xsd";

    public static final String ABSTRACT_CDKJBI_XSD = "abstractCDKjbi.xsd";

    /**
     * Extensions schema
     */
    public static final String CDKEXTENSIONS_XSD = "CDKextensions.xsd";

    /**
     * The schema factory
     */
    private static SchemaFactory factory;

    /**
     * <p>The JAXB unique unmarshaller.</p><p><b>MUST be used only through {@link #getUnMarshaller()} to be correctly initialized.</b></p>
     */
    private static Unmarshaller unmarshaller;

    /**
     * As the {@link Unmarshaller} is not thread-safe, it will be synchronized by this object.
     */
    private static Object unmarshallerSync;

    /**
     * The exception raised during JAXB unique unmarshaller creation.
     */
    private static CDKJBIDescriptorException unmarshCreationEx;

    /**
     * <p>The JAXB unique marshaller.</p><p><b>MUST be used only through {@link #getMarshaller()} to be correctly initialized.</b></p>
     */
    private static Marshaller marshaller;

    /**
     * As the {@link Marshaller} is not thread-safe, it will be synchronized by this object.
     */
    private static Object marshallerSync;

    /**
     * The exception raised during JAXB unique marshaller creation.
     */
    private static CDKJBIDescriptorException marshallerCreationEx;

    /**
     * <p>The JAXB unique validator.</p><p><b>MUST be used only through {@link #getValidator()} to be correctly initialized.</b></p>
     */
    private static Validator validator;

    /**
     * As the {@link Validator} is not thread-safe, it will be synchronized by this object.
     */
    private static Object validatorSync;

    /**
     * Static object initialization
     */
    static {
        init();
    }

    private static final void init() {

        factory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

        final URL cdkSchemaUrl = JBIDescriptorBuilder.class.getResource("/" + CDKJBI_XSD);
        final URL cdkExtUrl = JBIDescriptorBuilder.class.getResource("/" + CDKEXTENSIONS_XSD);

        try {
            InputStream cdkSchemaInputStream = cdkExtUrl.openStream();
            InputStream cdkExtInputStream = cdkSchemaUrl.openStream();

            final Schema schema = factory.newSchema(new StreamSource[] {
                    new StreamSource(cdkSchemaInputStream), new StreamSource(cdkExtInputStream) });
            
            // Validator initialization
            validatorSync = new Object();
            validator = schema.newValidator();

            final JAXBContext jaxbContext = JAXBContext.newInstance(new Class[] { Jbi.class });

            // Unmarshaller initialization
            unmarshallerSync = new Object();
            try {
                unmarshaller = jaxbContext.createUnmarshaller();
                unmarshaller.setSchema(schema);
            } catch (final JAXBException e) {
                unmarshCreationEx = new CDKJBIDescriptorException(
                        "Failed to create the JAXB unmarshaller", e);
            }

            // Marshaller initialization
            marshallerSync = new Object();
            try {
                marshaller = jaxbContext.createMarshaller();
                marshaller.setSchema(schema);
                marshaller.setProperty(Marshaller.JAXB_ENCODING, "UTF-8");
                marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
            } catch (JAXBException e) {
                marshallerCreationEx = new CDKJBIDescriptorException(
                        "Failed to create the JAXB marshaller", e);
            }

            cdkSchemaInputStream.close();
            cdkExtInputStream.close();

        } catch (final Exception e) {
            unmarshCreationEx = new CDKJBIDescriptorException(
                    "Failed to create the JAXB unmarshaller", e);
            marshallerCreationEx = new CDKJBIDescriptorException(
                    "Failed to create the JAXB marshaller and unmarshaller", e);
        }
    }

    /**
     * Free singleton resources. If the {@link JBIDescriptorBuilder} is used
     * after to be freed, resources will be re-allocated.
     */
    public final static void close() {
        synchronized (marshallerSync) {
            marshaller = null;
            marshallerCreationEx = null;
        }
        marshallerSync = null;

        synchronized (unmarshallerSync) {
            unmarshaller = null;
            unmarshCreationEx = null;
        }
        unmarshallerSync = null;
        
        synchronized (validatorSync) {
            validator = null;
        }
        validatorSync = null;

        factory = null;
    }

    /**
     * Validate the given JBI descriptor against the CDK JBI schema.
     * 
     * @param jbiDescriptorStream
     * @throws CDKJBIDescriptorException
     */
    public static void validateJBIDescriptor(final InputStream jbiDescriptorStream)
            throws CDKJBIDescriptorException {

        final Validator _validator = getValidator();
        // The validator is not thread safe
        synchronized (validatorSync) {
            try {
                _validator.validate(new StreamSource(jbiDescriptorStream));
            } catch (final SAXException e) {
                throw new CDKJBIDescriptorException(
                        "Failed to validate JBI descriptor against JBI schema", e);
            } catch (final IOException e) {
                throw new CDKJBIDescriptorException(
                        "Failed to validate JBI descriptor against JBI schema", e);
            }
        }
    }

    /**
     * Validate the given JBI descriptor against the component implementation
     * JBI schemas.
     * 
     * @param jbiDescriptorStream
     * @throws CDKJBIDescriptorException
     * 
     *             FIXME: The structure of XSDs with the redefine feature
     *             doesn't work, maybe a Xerces bug?
     */
    public static void validateJBIDescriptor(final InputStream jbiDescriptorStream,
            final String componentName) throws CDKJBIDescriptorException {
        final URL cdkExtUrl = JBIDescriptorBuilder.class.getResource("/" + CDKEXTENSIONS_XSD);
        final URL componentExtUrl = JBIDescriptorBuilder.class.getResource("/" + componentName
                + "extensions.xsd");
        final URL cdkSchemaUrl = JBIDescriptorBuilder.class.getResource("/" + ABSTRACT_CDKJBI_XSD);
        final URL componentSchemaUrl = JBIDescriptorBuilder.class.getResource("/" + componentName
                + "jbi.xsd");

        try {
            final Schema schema = getFactory().newSchema(
                    new StreamSource[] {
                    new StreamSource(cdkExtUrl.openStream()),
                    new StreamSource(componentExtUrl.openStream()),
                    new StreamSource(cdkSchemaUrl.openStream()),
                    new StreamSource(componentSchemaUrl.openStream()) });

            schema.newValidator().validate(new StreamSource(jbiDescriptorStream));

        } catch (SAXException e) {
            throw new CDKJBIDescriptorException(
                    "Failed to create the CDK component specific schema validation", e);
        } catch (IOException e) {
            throw new CDKJBIDescriptorException(
                    "Failed to create the CDK component specific schema validation", e);
        }
    }

    /**
     * <p>
     * Build the Jbi Java binding from the given JBI descriptor in XML Stream
     * form.
     * </p>
     * <p>
     * <b>Note</b> : if the stream is initialized from a {@link String}, prefer
     * to use
     * 
     * @param jbiDescriptorStream
     *            the XML JBI descriptor stream
     * @return the Jbi instance representing the JBI descriptor
     * @throws CDKJBIDescriptorException
     *             impossible to create the Java bindings
     */
    public static Jbi buildJavaJBIDescriptor(final InputStream jbiDescriptorStream)
            throws CDKJBIDescriptorException {
        try {
            final JAXBElement<Jbi> root;
            final Unmarshaller _unmarshaller = getUnMarshaller();
            // The unmarshaller is not thread safe
            synchronized (unmarshallerSync) {
                root = _unmarshaller
                        .unmarshal(new StreamSource(jbiDescriptorStream), Jbi.class);
            }

            return root.getValue();
        } catch (final JAXBException e) {
            throw new CDKJBIDescriptorException("Failed to build jbi descriptor", e);
        }
    }

    /**
     * <p>
     * Build the Jbi Java binding from the given JBI descriptor in XML character
     * reader form.
     * </p>
     * <p>
     * <b>Note</b> : <b> MUST </b> be used to parse a JBI descriptor available
     * as {@link String}.
     * 
     * @param jbiDescriptorReader
     *            the XML JBI descriptor character reader
     * @return the Jbi instance representing the JBI descriptor
     * @throws CDKJBIDescriptorException
     *             impossible to create the Java bindings
     */
    public static Jbi buildJavaJBIDescriptor(final Reader jbiDescriptorReader)
            throws CDKJBIDescriptorException {
        try {
            final JAXBElement<Jbi> root;
            final Unmarshaller _unmarshaller = getUnMarshaller();
            // The unmarshaller is not thread safe
            synchronized (unmarshallerSync) {
                root = _unmarshaller
                        .unmarshal(new StreamSource(jbiDescriptorReader), Jbi.class);
            }

            return root.getValue();
        } catch (final JAXBException e) {
            throw new CDKJBIDescriptorException("Failed to build jbi descriptor", e);
        }
    }

    
    /**
     * Build the XML nodes tree from the JBI descriptor in Java classes form.
     * 
     * @param jbiDescriptorClass
     *            The JBI Descriptor root class
     * @param jbiDescriptorNode
     *            The XML Node to fill with the JBI descriptor XML nodes
     * @throws JBIDescriptorException
     *             The exception raised during the marshaller creation or the
     *             exception raised during the build of the XML nodes.
     */
    public static void buildXmlJBIdescriptor(final Jbi jbiDescriptorClass, Node jbiDescriptorNode)
            throws CDKJBIDescriptorException {

        try {
            final Marshaller _marshaller = getMarshaller();
            // The marshaller is not thread safe
            synchronized (marshallerSync) {
                _marshaller.marshal(jbiDescriptorClass, jbiDescriptorNode);
            }
        } catch (JAXBException e) {
            throw new CDKJBIDescriptorException(
                    "Failed to build XML binding from JBI descriptor Java classes", e);
        }
    }

    /**
     * Build the XML nodes tree from the JBI descriptor in Java classes form. A
     * default Document id built-in and filled with the created nodes.
     * 
     * @param jbiDescriptorClass
     *            The JBI Descriptor root class
     * @return The built-in @ code Document} filled with the generated nodes
     * @throws JBIDescriptorException
     *             The exception raised during the marshaller creation or the
     *             exception raised during the build of the XML nodes.
     */
    public static Document buildXmlJBIdescriptor(final Jbi jbiDescriptorClass)
            throws CDKJBIDescriptorException {

        Document document = DocumentBuilders.newDocument();

        try {
            final Marshaller _marshaller = getMarshaller();
            // The marshaller is not thread safe
            synchronized (marshallerSync) {
                _marshaller.marshal(jbiDescriptorClass, document);
            }
        } catch (JAXBException e) {
            throw new CDKJBIDescriptorException(
                    "Failed to build XML binding from JBI descriptor Java classes", e);
        }

        return document;
    }

    /**
     * Build the output Stream from the JBI descriptor in Java classes form.
     * 
     * @param jbiDescriptorClass
     *            The JBI Descriptor root class
     * @param jbiDescriptorNode
     *            The output Stream to fill with the JBI descriptor
     * @throws JBIDescriptorException
     *             The exception raised during the marshaller creation or the
     *             exception raised during the build of the XML nodes.
     */
    public static void buildXmlJBIdescriptor(final Jbi jbiDescriptorClass, OutputStream ouputStream)
            throws CDKJBIDescriptorException {

        try {
            final Marshaller _marshaller = getMarshaller();
            // The marshaller is not thread safe
            synchronized (marshallerSync) {
                _marshaller.marshal(jbiDescriptorClass, ouputStream);
            }
        } catch (JAXBException e) {
            throw new CDKJBIDescriptorException(
                    "Failed to build XML binding from JBI descriptor Java classes", e);
        }
    }

    /**
     * Get the unmarshaller instance, initializes it if needed.
     * 
     * @return the unmarshaller instance
     * @throws CDKJBIDescriptorException
     *             The exception raised during the unmarshaller creation.
     */
    private final static synchronized Unmarshaller getUnMarshaller()
            throws CDKJBIDescriptorException {

        if (unmarshaller == null && unmarshCreationEx == null) {
            // Initialization of the unmarshaller is needed
            init();
        }

        if (unmarshCreationEx != null) {
            throw unmarshCreationEx;
        }

        return unmarshaller;
    }


    /**
     * Get the marshaller instance, initializes it if needed.
     * 
     * @return the marshaller instance
     * @throws JBIDescriptorException
     *             The exception raised during the marshaller creation.
     */
    private final static synchronized Marshaller getMarshaller() throws CDKJBIDescriptorException {

        if (marshaller == null && marshallerCreationEx == null) {
            // Initialization of the marshaller is needed
            init();
        }

        if (marshallerCreationEx != null) {
            throw marshallerCreationEx;
        }

        return marshaller;
    }

    /**
     * Get the factory instance, initializes it if needed.
     * 
     * @return the factory instance
     */
    private final static synchronized SchemaFactory getFactory() {

        if (factory == null) {
            // Initialization of the factory is needed
            init();
        }

        return factory;
    }

    /**
     * Get the validator instance, initializes it if needed.
     * 
     * @return the validator instance
     */
    private final static synchronized Validator getValidator() {

        if (validator == null) {
            // Initialization of the validator is needed
            init();
        }

        return validator;
    }
}
